@@ -2,21 +2,19 @@
  * adder.c
  * Copyright 2011 John Lindgren
  *
- * This file is part of Audacious.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * Audacious is free software: you can redistribute it and/or modify it under
- * the terms of the GNU General Public License as published by the Free Software
- * Foundation, version 2 or version 3 of the License.
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions, and the following disclaimer.
  *
- * Audacious is distributed in the hope that it will be useful, but WITHOUT ANY
- * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
- * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
+ * 2. Redistributions in binary form must reproduce the above copyright notice,
+ *    this list of conditions, and the following disclaimer in the documentation
+ *    provided with the distribution.
  *
- * You should have received a copy of the GNU General Public License along with
- * Audacious. If not, see <http://www.gnu.org/licenses/>.
- *
- * The Audacious team does not consider modular code linking to Audacious or
- * using our public API to be a derived work.
+ * This software is provided "as is" and without any warranty, express or
+ * implied. In no event shall the authors be liable for any damages arising from
+ * the use of this software.
  */
 
 #include <dirent.h>
@@ -29,9 +27,8 @@
 #include <libaudcore/audstrings.h>
 #include <libaudcore/hook.h>
 
-#include "config.h"
+#include "drct.h"
 #include "i18n.h"
-#include "playback.h"
 #include "playlist.h"
 #include "plugins.h"
 #include "main.h"
@@ -78,27 +75,19 @@ static bool_t status_cb (void * unused)
         gtk_window_set_resizable ((GtkWindow *) status_window, FALSE);
         gtk_container_set_border_width ((GtkContainer *) status_window, 6);
 
-        GtkWidget * vbox = gtk_vbox_new (FALSE, 6);
+        GtkWidget * vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
         gtk_container_add ((GtkContainer *) status_window, vbox);
 
         status_path_label = gtk_label_new (NULL);
-#if GTK_CHECK_VERSION (3, 0, 0)
         gtk_label_set_width_chars ((GtkLabel *) status_path_label, 40);
         gtk_label_set_max_width_chars ((GtkLabel *) status_path_label, 40);
-#else
-        gtk_widget_set_size_request (status_path_label, 320, -1);
-#endif
         gtk_label_set_ellipsize ((GtkLabel *) status_path_label,
          PANGO_ELLIPSIZE_MIDDLE);
         gtk_box_pack_start ((GtkBox *) vbox, status_path_label, FALSE, FALSE, 0);
 
         status_count_label = gtk_label_new (NULL);
-#if GTK_CHECK_VERSION (3, 0, 0)
         gtk_label_set_width_chars ((GtkLabel *) status_count_label, 40);
         gtk_label_set_max_width_chars ((GtkLabel *) status_count_label, 40);
-#else
-        gtk_widget_set_size_request (status_count_label, 320, -1);
-#endif
         gtk_box_pack_start ((GtkBox *) vbox, status_count_label, FALSE, FALSE, 0);
 
         gtk_widget_show_all (status_window);
@@ -418,11 +407,10 @@ static bool_t add_finish (void * unused)
 
         if (result->play && playlist_entry_count (playlist) > count)
         {
-            playlist_set_playing (playlist);
             if (! get_bool (NULL, "shuffle"))
                 playlist_set_position (playlist, result->at);
 
-            playback_play (0, FALSE);
+            drct_play_playlist (playlist);
         }
 
     FREE:
@@ -511,6 +499,11 @@ void adder_cleanup (void)
     pthread_mutex_unlock (& mutex);
     pthread_join (add_thread, NULL);
 
+    g_list_free_full (add_tasks, (GDestroyNotify) add_task_free);
+    add_tasks = NULL;
+    g_list_free_full (add_results, (GDestroyNotify) add_result_free);
+    add_results = NULL;
+
     if (add_source)
     {
         g_source_remove (add_source);
@@ -554,23 +547,30 @@ void playlist_entry_insert_filtered (int playlist, int at,
 
 bool_t playlist_add_in_progress (int playlist)
 {
-    int playlist_id = playlist_get_unique_id (playlist);
-    g_return_val_if_fail (playlist_id >= 0, FALSE);
-
     pthread_mutex_lock (& mutex);
 
-    for (GList * node = add_tasks; node; node = node->next)
+    if (playlist >= 0)
     {
-        if (((AddTask *) node->data)->playlist_id == playlist_id)
+        int playlist_id = playlist_get_unique_id (playlist);
+
+        for (GList * node = add_tasks; node; node = node->next)
+        {
+            if (((AddTask *) node->data)->playlist_id == playlist_id)
+                goto YES;
+        }
+
+        if (current_playlist_id == playlist_id)
             goto YES;
+
+        for (GList * node = add_results; node; node = node->next)
+        {
+            if (((AddResult *) node->data)->playlist_id == playlist_id)
+                goto YES;
+        }
     }
-
-    if (current_playlist_id == playlist_id)
-        goto YES;
-
-    for (GList * node = add_results; node; node = node->next)
+    else
     {
-        if (((AddResult *) node->data)->playlist_id == playlist_id)
+        if (add_tasks || current_playlist_id >= 0 || add_results)
             goto YES;
     }
 
