@@ -37,6 +37,7 @@ static const char * const core_defaults[] = {
  "clear_playlist", "TRUE",
  "open_to_temporary", "TRUE",
  "resume_playback_on_startup", "FALSE",
+ "show_interface", "TRUE",
 
  /* equalizer */
  "eqpreset_default_file", "",
@@ -80,6 +81,11 @@ static const char * const core_defaults[] = {
  "stop_after_current_song", "FALSE",
 
  /* playlist */
+#ifdef _WIN32
+ "convert_backslash", "TRUE",
+#else
+ "convert_backslash", "FALSE",
+#endif
  "generic_title_format", "${?artist:${artist} - }${?album:${album} - }${title}",
  "leading_zero", "FALSE",
  "metadata_on_play", "FALSE",
