@@ -24,7 +24,7 @@
 #include <libaudcore/audstrings.h>
 #include <libaudcore/hook.h>
 
-#include "config.h"
+#include "drct.h"
 #include "i18n.h"
 #include "interface.h"
 #include "misc.h"
@@ -33,8 +33,6 @@
 #include "playlist.h"
 #include "plugin.h"
 
-static void playback_start (int playlist, int entry, int seek_time, bool_t pause);
-
 static const struct OutputAPI output_api = {
  .open_audio = output_open_audio,
  .set_replaygain_info = output_set_replaygain_info,
@@ -46,31 +44,38 @@ static const struct OutputAPI output_api = {
 
 static InputPlayback playback_api;
 
-static bool_t playing = FALSE;
-static bool_t playback_error;
-static int failed_entries;
-
-static char * current_filename; /* pooled */
-
-static int current_entry;
-static char * current_title; /* pooled */
-static int current_length;
-
-static InputPlugin * current_decoder;
-static void * current_data;
-static int current_bitrate, current_samplerate, current_channels;
-
-static ReplayGainInfo gain_from_playlist;
-
-static int time_offset, initial_seek;
-static bool_t paused;
-
 static pthread_t playback_thread_handle;
 static int end_source = 0;
 
 static pthread_mutex_t ready_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t ready_cond = PTHREAD_COND_INITIALIZER;
-static bool_t ready_flag;
+
+/* level 1 data (persists to end of song) */
+static bool_t playing = FALSE;
+static bool_t restart_flag = FALSE;
+static int time_offset = 0, initial_seek = 0;
+static bool_t paused = FALSE;
+static bool_t ready_flag = FALSE;
+static bool_t playback_error = FALSE;
+
+static void * current_data = NULL;
+static int current_bitrate = -1, current_samplerate = -1, current_channels = -1;
+
+/* level 2 data (persists when restarting same song) */
+static int current_entry = -1;
+static char * current_filename = NULL; /* pooled */
+static char * current_title = NULL; /* pooled */
+static int current_length = -1;
+
+static InputPlugin * current_decoder = NULL;
+static VFSFile * current_file = NULL;
+static ReplayGainInfo gain_from_playlist;
+
+static int repeat_a = -1, repeat_b = -1;
+
+/* level 3 data (persists to end of playlist) */
+static bool_t stopped = TRUE;
+static int failed_entries = 0;
 
 /* clears gain info if tuple == NULL */
 static void read_gain_from_tuple (const Tuple * tuple)
@@ -106,7 +111,8 @@ static bool_t update_from_playlist (void)
     char * title = playback_entry_get_title ();
     int length = playback_entry_get_length ();
 
-    if (entry == current_entry && ! g_strcmp0 (title, current_title) && length == current_length)
+    /* pointer comparison works for pooled strings */
+    if (entry == current_entry && title == current_title && length == current_length)
     {
         str_unref (title);
         return FALSE;
@@ -119,11 +125,16 @@ static bool_t update_from_playlist (void)
     return TRUE;
 }
 
-bool_t playback_get_ready (void)
+bool_t drct_get_ready (void)
 {
-    g_return_val_if_fail (playing, FALSE);
+    if (! playing)
+        return FALSE;
+
     pthread_mutex_lock (& ready_mutex);
-    bool_t ready = ready_flag;
+
+    /* on restart, always report ready */
+    bool_t ready = ready_flag || restart_flag;
+
     pthread_mutex_unlock (& ready_mutex);
     return ready;
 }
@@ -133,13 +144,17 @@ static void set_pb_ready (InputPlayback * p)
     g_return_if_fail (playing);
     pthread_mutex_lock (& ready_mutex);
 
-    update_from_playlist ();
+    /* on restart, don't update or send "playback ready" */
+    if (! restart_flag)
+    {
+        update_from_playlist ();
+        event_queue ("playback ready", NULL);
+    }
+
     ready_flag = TRUE;
 
     pthread_cond_signal (& ready_cond);
     pthread_mutex_unlock (& ready_mutex);
-
-    event_queue ("playback ready", NULL);
 }
 
 static void wait_until_ready (void)
@@ -147,6 +162,7 @@ static void wait_until_ready (void)
     g_return_if_fail (playing);
     pthread_mutex_lock (& ready_mutex);
 
+    /* on restart, we still have to wait, but presumably not long */
     while (! ready_flag)
         pthread_cond_wait (& ready_cond, & ready_mutex);
 
@@ -157,16 +173,18 @@ static void update_cb (void * hook_data, void * user_data)
 {
     g_return_if_fail (playing);
 
-    if (GPOINTER_TO_INT (hook_data) < PLAYLIST_UPDATE_METADATA || ! playback_get_ready ())
+    if (GPOINTER_TO_INT (hook_data) < PLAYLIST_UPDATE_METADATA || ! drct_get_ready ())
         return;
 
     if (update_from_playlist ())
         event_queue ("title change", NULL);
 }
 
-int playback_get_time (void)
+int drct_get_time (void)
 {
-    g_return_val_if_fail (playing, 0);
+    if (! playing)
+        return 0;
+
     wait_until_ready ();
 
     int time = -1;
@@ -180,25 +198,11 @@ int playback_get_time (void)
     return time - time_offset;
 }
 
-void playback_play (int seek_time, bool_t pause)
+void drct_pause (void)
 {
-    g_return_if_fail (! playing);
-
-    int playlist = playlist_get_playing ();
-    if (playlist < 0)
+    if (! playing)
         return;
 
-    int entry = playlist_get_position (playlist);
-    if (entry < 0)
-        return;
-
-    failed_entries = 0;
-    playback_start (playlist, entry, seek_time, pause);
-}
-
-void playback_pause (void)
-{
-    g_return_if_fail (playing);
     wait_until_ready ();
 
     if (! current_decoder || ! current_decoder->pause)
@@ -213,40 +217,7 @@ void playback_pause (void)
         hook_call ("playback unpause", NULL);
 }
 
-static void playback_cleanup (void)
-{
-    g_return_if_fail (playing);
-
-    pthread_join (playback_thread_handle, NULL);
-    playing = FALSE;
-
-    event_queue_cancel ("playback ready", NULL);
-    event_queue_cancel ("playback seek", NULL);
-    event_queue_cancel ("info change", NULL);
-    event_queue_cancel ("title change", NULL);
-
-    if (end_source)
-    {
-        g_source_remove (end_source);
-        end_source = 0;
-    }
-
-    str_unref (current_filename);
-    current_filename = NULL;
-    str_unref (current_title);
-    current_title = NULL;
-
-    hook_dissociate ("playlist update", update_cb);
-}
-
-static void complete_stop (void)
-{
-    output_drain ();
-    hook_call ("playback stop", NULL);
-    set_bool (NULL, "stop_after_current_song", FALSE);
-}
-
-void playback_stop (void)
+static void playback_finish (void)
 {
     g_return_if_fail (playing);
     wait_until_ready ();
@@ -254,8 +225,77 @@ void playback_stop (void)
     if (current_decoder)
         current_decoder->stop (& playback_api);
 
-    playback_cleanup ();
-    complete_stop ();
+    pthread_join (playback_thread_handle, NULL);
+    output_close_audio ();
+
+    hook_dissociate ("playlist update", update_cb);
+
+    if (end_source)
+    {
+        g_source_remove (end_source);
+        end_source = 0;
+    }
+
+    /* level 1 data cleanup */
+    playing = FALSE;
+    restart_flag = FALSE;
+    time_offset = initial_seek = 0;
+    paused = FALSE;
+    ready_flag = FALSE;
+    playback_error = FALSE;
+
+    current_data = NULL;
+    current_bitrate = current_samplerate = current_channels = -1;
+}
+
+static void playback_cleanup (void)
+{
+    g_return_if_fail (current_filename);
+    playback_finish ();
+
+    event_queue_cancel ("playback ready", NULL);
+    event_queue_cancel ("playback seek", NULL);
+    event_queue_cancel ("info change", NULL);
+    event_queue_cancel ("title change", NULL);
+
+    set_bool (NULL, "stop_after_current_song", FALSE);
+
+    /* level 2 data cleanup */
+    current_entry = -1;
+    str_unref (current_filename);
+    current_filename = NULL;
+    str_unref (current_title);
+    current_title = NULL;
+    current_length = -1;
+
+    current_decoder = NULL;
+
+    if (current_file)
+    {
+        vfs_fclose (current_file);
+        current_file = NULL;
+    }
+
+    read_gain_from_tuple (NULL);
+
+    repeat_a = repeat_b = -1;
+}
+
+void playback_stop (void)
+{
+    if (stopped)
+        return;
+
+    if (current_filename)
+        playback_cleanup ();
+
+    output_drain ();
+
+    /* level 3 data cleanup */
+    stopped = TRUE;
+    failed_entries = 0;
+
+    hook_call ("playback stop", NULL);
 }
 
 static bool_t end_cb (void * unused)
@@ -269,62 +309,92 @@ static bool_t end_cb (void * unused)
     else
         failed_entries = 0;
 
-    playback_cleanup ();
-
     int playlist = playlist_get_playing ();
-    bool_t play;
-
-    if (get_bool (NULL, "no_playlist_advance"))
-        play = get_bool (NULL, "repeat") && ! failed_entries;
-    else if (! (play = playlist_next_song (playlist, get_bool (NULL, "repeat"))))
-        playlist_set_position (playlist, -1);
-    else if (failed_entries >= 10)
-        play = FALSE;
 
     if (get_bool (NULL, "stop_after_current_song"))
-        play = FALSE;
+        goto STOP;
 
-    if (play)
-        playback_start (playlist, playlist_get_position (playlist), 0, FALSE);
+    if (repeat_a >= 0 || repeat_b >= 0)
+    {
+        if (failed_entries)
+            goto STOP;
+
+        playback_play (MAX (repeat_a, 0), FALSE);
+    }
+    else if (get_bool (NULL, "no_playlist_advance"))
+    {
+        if (failed_entries || ! get_bool (NULL, "repeat"))
+            goto STOP;
+
+        playback_play (0, FALSE);
+    }
     else
     {
-        complete_stop ();
-        hook_call ("playlist end reached", NULL);
+        if (failed_entries >= 10)
+            goto STOP;
+
+        if (! playlist_next_song (playlist, get_bool (NULL, "repeat")))
+        {
+            playlist_set_position (playlist, -1);
+            hook_call ("playlist end reached", NULL);
+        }
     }
 
     return FALSE;
+
+STOP:
+    /* stop playback and set position to beginning of song */
+    playlist_set_playing (-1);
+    playlist_set_position (playlist, playlist_get_position (playlist));
+    return FALSE;
 }
 
 static void * playback_thread (void * unused)
 {
-    PluginHandle * p = playback_entry_get_decoder ();
-    current_decoder = p ? plugin_get_header (p) : NULL;
-
     if (! current_decoder)
     {
-        SPRINTF (error, _("No decoder found for %s."), current_filename);
-        interface_show_error (error);
-        playback_error = TRUE;
-        goto DONE;
-    }
+        PluginHandle * p = playback_entry_get_decoder ();
+        current_decoder = p ? plugin_get_header (p) : NULL;
 
-    current_data = NULL;
-    current_bitrate = 0;
-    current_samplerate = 0;
-    current_channels = 0;
+        if (! current_decoder)
+        {
+            SPRINTF (error, _("No decoder found for %s."), current_filename);
+            interface_show_error (error);
+            playback_error = TRUE;
+            goto DONE;
+        }
+    }
 
     Tuple * tuple = playback_entry_get_tuple ();
     read_gain_from_tuple (tuple);
+
+    int start_time = time_offset = 0;
+    int end_time = -1;
+
+    if (tuple && playback_entry_get_length () > 0)
+    {
+        if (tuple_get_value_type (tuple, FIELD_SEGMENT_START, NULL) == TUPLE_INT)
+            time_offset = tuple_get_int (tuple, FIELD_SEGMENT_START, NULL);
+
+        start_time = time_offset + MAX (initial_seek, 0);
+
+        if (repeat_b >= 0)
+            end_time = time_offset + repeat_b;
+        else if (tuple_get_value_type (tuple, FIELD_SEGMENT_END, NULL) == TUPLE_INT)
+            end_time = tuple_get_int (tuple, FIELD_SEGMENT_END, NULL);
+    }
+
     if (tuple)
         tuple_unref (tuple);
 
-    bool_t seekable = (playback_entry_get_length () > 0);
-
-    VFSFile * file = NULL;
-
     if (! current_decoder->schemes || ! current_decoder->schemes[0])
     {
-        if (! (file = vfs_fopen (current_filename, "r")))
+        if (current_file)
+            vfs_rewind (current_file);
+        else
+            current_file = vfs_fopen (current_filename, "r");
+
+        if (! current_file)
         {
             SPRINTF (error, _("%s could not be opened."), current_filename);
             interface_show_error (error);
@@ -333,15 +403,8 @@ static void * playback_thread (void * unused)
         }
     }
 
-    time_offset = seekable ? playback_entry_get_start_time () : 0;
     playback_error = ! current_decoder->play (& playback_api, current_filename,
-     file, seekable ? time_offset + initial_seek : 0,
-     seekable ? playback_entry_get_end_time () : -1, paused);
-
-    output_close_audio ();
-
-    if (file)
-        vfs_fclose (file);
+     current_file, start_time, end_time, paused);
 
 DONE:
     if (! ready_flag)
@@ -351,44 +414,58 @@ DONE:
     return NULL;
 }
 
-static void playback_start (int playlist, int entry, int seek_time, bool_t pause)
+void playback_play (int seek_time, bool_t pause)
 {
-    g_return_if_fail (! playing);
+    char * new_filename = playback_entry_get_filename ();
+    g_return_if_fail (new_filename);
 
-    current_filename = playlist_entry_get_filename (playlist, entry);
-    g_return_if_fail (current_filename);
+    /* pointer comparison works for pooled strings */
+    if (new_filename == current_filename)
+    {
+        if (playing)
+            playback_finish ();
+
+        str_unref (new_filename);
+        restart_flag = TRUE;
+    }
+    else
+    {
+        if (current_filename)
+            playback_cleanup ();
+
+        current_filename = new_filename;
+    }
 
     playing = TRUE;
-    playback_error = FALSE;
-    ready_flag = FALSE;
-
-    current_entry = -1;
-    current_title = NULL;
-    current_length = 0;
-
     initial_seek = seek_time;
     paused = pause;
+    stopped = FALSE;
 
     hook_associate ("playlist update", update_cb, NULL);
     pthread_create (& playback_thread_handle, NULL, playback_thread, NULL);
 
-    hook_call ("playback begin", NULL);
+    /* on restart, send "playback seek" instead of "playback begin" */
+    if (restart_flag)
+        hook_call ("playback seek", NULL);
+    else
+        hook_call ("playback begin", NULL);
 }
 
-bool_t playback_get_playing (void)
+bool_t drct_get_playing (void)
 {
     return playing;
 }
 
-bool_t playback_get_paused (void)
+bool_t drct_get_paused (void)
 {
-    g_return_val_if_fail (playing, FALSE);
     return paused;
 }
 
-void playback_seek (int time)
+void drct_seek (int time)
 {
-    g_return_if_fail (playing);
+    if (! playing)
+        return;
+
     wait_until_ready ();
 
     if (! current_decoder || ! current_decoder->mseek || current_length < 1)
@@ -426,7 +503,7 @@ static void set_params (InputPlayback * p, int bitrate, int samplerate,
     current_samplerate = samplerate;
     current_channels = channels;
 
-    if (playback_get_ready ())
+    if (drct_get_ready ())
         event_queue ("info change", NULL);
 }
 
@@ -453,15 +530,19 @@ static InputPlayback playback_api = {
     .set_gain_from_playlist = set_gain_from_playlist,
 };
 
-char * playback_get_filename (void)
+char * drct_get_filename (void)
 {
-    g_return_val_if_fail (playing, NULL);
+    if (! playing)
+        return NULL;
+
     return str_ref (current_filename);
 }
 
-char * playback_get_title (void)
+char * drct_get_title (void)
 {
-    g_return_val_if_fail (playing, NULL);
+    if (! playing)
+        return NULL;
+
     wait_until_ready ();
 
     char s[32];
@@ -481,44 +562,82 @@ char * playback_get_title (void)
         s[0] = 0;
 
     if (get_bool (NULL, "show_numbers_in_pl"))
-        return str_printf ("%d. %s%s", 1 + playlist_get_position
-         (playlist_get_playing ()), current_title, s);
+        return str_printf ("%d. %s%s", 1 + current_entry, current_title, s);
 
     return str_printf ("%s%s", current_title, s);
 }
 
-int playback_get_length (void)
+int drct_get_length (void)
 {
-    g_return_val_if_fail (playing, 0);
-    wait_until_ready ();
+    if (playing)
+        wait_until_ready ();
 
     return current_length;
 }
 
-void playback_get_info (int * bitrate, int * samplerate, int * channels)
+void drct_get_info (int * bitrate, int * samplerate, int * channels)
 {
-    g_return_if_fail (playing);
-    wait_until_ready ();
+    if (playing)
+        wait_until_ready ();
 
     * bitrate = current_bitrate;
     * samplerate = current_samplerate;
     * channels = current_channels;
 }
 
-void playback_get_volume (int * l, int * r)
+void drct_get_volume (int * l, int * r)
 {
-    if (playing && playback_get_ready () && current_decoder &&
+    if (playing && drct_get_ready () && current_decoder &&
      current_decoder->get_volume && current_decoder->get_volume (l, r))
         return;
 
     output_get_volume (l, r);
 }
 
-void playback_set_volume (int l, int r)
+void drct_set_volume (int l, int r)
 {
-    if (playing && playback_get_ready () && current_decoder &&
+    l = CLAMP (l, 0, 100);
+    r = CLAMP (r, 0, 100);
+
+    if (playing && drct_get_ready () && current_decoder &&
      current_decoder->set_volume && current_decoder->set_volume (l, r))
         return;
 
     output_set_volume (l, r);
 }
+
+void drct_set_ab_repeat (int a, int b)
+{
+    if (! playing)
+        return;
+
+    wait_until_ready ();
+
+    if (current_length < 1)
+        return;
+
+    repeat_a = a;
+
+    if (repeat_b != b)
+    {
+        repeat_b = b;
+
+        /* Restart playback so the new setting takes effect.  We could add
+         * something like InputPlugin::set_stop_time(), but this is the only
+         * place it would be used. */
+        int seek_time = drct_get_time ();
+        bool_t was_paused = paused;
+
+        if (repeat_b >= 0 && seek_time >= repeat_b)
+            seek_time = MAX (repeat_a, 0);
+
+        playback_finish ();
+        playback_play (seek_time, was_paused);
+    }
+}
+
+void drct_get_ab_repeat (int * a, int * b)
+{
+    * a = playing ? repeat_a : -1;
+    * b = playing ? repeat_b : -1;
+}
