@@ -2,21 +2,19 @@
  * probe-buffer.c
  * Copyright 2010-2011 John Lindgren
  *
- * This file is part of Audacious.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * Audacious is free software: you can redistribute it and/or modify it under
- * the terms of the GNU General Public License as published by the Free Software
- * Foundation, version 2 or version 3 of the License.
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions, and the following disclaimer.
  *
- * Audacious is distributed in the hope that it will be useful, but WITHOUT ANY
- * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
- * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
+ * 2. Redistributions in binary form must reproduce the above copyright notice,
+ *    this list of conditions, and the following disclaimer in the documentation
+ *    provided with the distribution.
  *
- * You should have received a copy of the GNU General Public License along with
- * Audacious. If not, see <http://www.gnu.org/licenses/>.
- *
- * The Audacious team does not consider modular code linking to Audacious or
- * using our public API to be a derived work.
+ * This software is provided "as is" and without any warranty, express or
+ * implied. In no event shall the authors be liable for any damages arising from
+ * the use of this software.
  */
 
 #include <glib.h>
@@ -71,8 +69,7 @@ static int64_t probe_buffer_fread (void * buffer, int64_t size, int64_t count,
 static int64_t probe_buffer_fwrite (const void * data, int64_t size, int64_t count,
  VFSFile * file)
 {
-    /* not implemented */
-    return 0;
+    return 0; /* not allowed */
 }
 
 static int probe_buffer_getc (VFSFile * file)
@@ -86,7 +83,7 @@ static int probe_buffer_fseek (VFSFile * file, int64_t offset, int whence)
     ProbeBuffer * p = vfs_get_handle (file);
 
     if (whence == SEEK_END)
-        return -1;
+        return -1; /* not allowed */
 
     if (whence == SEEK_CUR)
         offset += p->at;
@@ -119,18 +116,26 @@ static int64_t probe_buffer_ftell (VFSFile * file)
 static bool_t probe_buffer_feof (VFSFile * file)
 {
     ProbeBuffer * p = vfs_get_handle (file);
-    return (p->at < p->filled) ? FALSE : vfs_feof (p->file);
+
+    if (p->at < p->filled)
+        return FALSE;
+    if (p->at == sizeof p->buffer)
+        return TRUE;
+
+    return vfs_feof (p->file);
 }
 
 static int probe_buffer_ftruncate (VFSFile * file, int64_t size)
 {
-    /* not implemented */
-    return -1;
+    return -1; /* not allowed */
 }
 
 static int64_t probe_buffer_fsize (VFSFile * file)
 {
-    return vfs_fsize (((ProbeBuffer *) vfs_get_handle (file))->file);
+    ProbeBuffer * p = vfs_get_handle (file);
+
+    int64_t size = vfs_fsize (p->file);
+    return MIN (size, sizeof p->buffer);
 }
 
 static char * probe_buffer_get_metadata (VFSFile * file, const char * field)
@@ -140,19 +145,19 @@ static char * probe_buffer_get_metadata (VFSFile * file, const char * field)
 
 static VFSConstructor probe_buffer_table =
 {
-	.vfs_fopen_impl = NULL,
-	.vfs_fclose_impl = probe_buffer_fclose,
-	.vfs_fread_impl = probe_buffer_fread,
-	.vfs_fwrite_impl = probe_buffer_fwrite,
-	.vfs_getc_impl = probe_buffer_getc,
-	.vfs_ungetc_impl = probe_buffer_ungetc,
-	.vfs_fseek_impl = probe_buffer_fseek,
-	.vfs_rewind_impl = probe_buffer_rewind,
-	.vfs_ftell_impl = probe_buffer_ftell,
-	.vfs_feof_impl = probe_buffer_feof,
-	.vfs_ftruncate_impl = probe_buffer_ftruncate,
-	.vfs_fsize_impl = probe_buffer_fsize,
-	.vfs_get_metadata_impl = probe_buffer_get_metadata,
+    .vfs_fopen_impl = NULL,
+    .vfs_fclose_impl = probe_buffer_fclose,
+    .vfs_fread_impl = probe_buffer_fread,
+    .vfs_fwrite_impl = probe_buffer_fwrite,
+    .vfs_getc_impl = probe_buffer_getc,
+    .vfs_ungetc_impl = probe_buffer_ungetc,
+    .vfs_fseek_impl = probe_buffer_fseek,
+    .vfs_rewind_impl = probe_buffer_rewind,
+    .vfs_ftell_impl = probe_buffer_ftell,
+    .vfs_feof_impl = probe_buffer_feof,
+    .vfs_ftruncate_impl = probe_buffer_ftruncate,
+    .vfs_fsize_impl = probe_buffer_fsize,
+    .vfs_get_metadata_impl = probe_buffer_get_metadata,
 };
 
 VFSFile * probe_buffer_new (const char * filename)
