@@ -1,22 +1,21 @@
 /*
- *  vfs.c
- *  Copyright 2006-2011 William Pitcock, Daniel Barkalow, Ralf Ertzinger,
- *                      Yoshiki Yazawa, Matti Hämäläinen, and John Lindgren
+ * vfs.c
+ * Copyright 2006-2011 William Pitcock, Daniel Barkalow, Ralf Ertzinger,
+ *                     Yoshiki Yazawa, Matti Hämäläinen, and John Lindgren
  *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; under version 3 of the License.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions, and the following disclaimer.
  *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses>.
+ * 2. Redistributions in binary form must reproduce the above copyright notice,
+ *    this list of conditions, and the following disclaimer in the documentation
+ *    provided with the distribution.
  *
- *  The Audacious team does not consider modular code linking to
- *  Audacious or using our public API to be a derived work.
+ * This software is provided "as is" and without any warranty, express or
+ * implied. In no event shall the authors be liable for any damages arising from
+ * the use of this software.
  */
 
 #include <glib.h>
@@ -30,8 +29,6 @@
 #include <sys/types.h>
 #include <string.h>
 
-#include "config.h"
-
 #define VFS_SIG ('V' | ('F' << 8) | ('S' << 16))
 
 /**
@@ -412,15 +409,40 @@ vfs_file_test(const char * path, int test)
         return FALSE; /* only local files are handled */
 
     char * path2 = uri_to_filename (path);
+    if (! path2)
+        return FALSE;
 
-    if (path2 == NULL)
-        path2 = g_strdup(path);
+#ifdef S_ISLNK
+    if (test & VFS_IS_SYMLINK)
+    {
+        struct stat st;
+        if (lstat (path2, & st) < 0)
+            return FALSE;
 
-    bool_t ret = g_file_test (path2, test);
+        if (S_ISLNK (st.st_mode))
+            test &= ~VFS_IS_SYMLINK;
+    }
+#endif
 
-    g_free(path2);
+    if (test & (VFS_IS_REGULAR | VFS_IS_DIR | VFS_IS_EXECUTABLE | VFS_EXISTS))
+    {
+        struct stat st;
+        if (stat (path2, & st) < 0)
+            return FALSE;
 
-    return ret;
+        if (S_ISREG (st.st_mode))
+            test &= ~VFS_IS_REGULAR;
+        if (S_ISDIR (st.st_mode))
+            test &= ~VFS_IS_DIR;
+        if (st.st_mode & S_IXUSR)
+            test &= ~VFS_IS_EXECUTABLE;
+
+        test &= ~VFS_EXISTS;
+    }
+
+    g_free (path2);
+
+    return ! test;
 }
 
 /**
