@@ -1,61 +1,40 @@
 /*
  * libaudgui/infowin.c
- * Copyright 2006-2011 William Pitcock, Tomasz Moń, Eugene Zagidullin, and
+ * Copyright 2006-2012 William Pitcock, Tomasz Moń, Eugene Zagidullin, and
  *                     John Lindgren
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; under version 3 of the License.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions, and the following disclaimer.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses>.
+ * 2. Redistributions in binary form must reproduce the above copyright notice,
+ *    this list of conditions, and the following disclaimer in the documentation
+ *    provided with the distribution.
  *
- * The Audacious team does not consider modular code linking to
- * Audacious or using our public API to be a derived work.
+ * This software is provided "as is" and without any warranty, express or
+ * implied. In no event shall the authors be liable for any damages arising from
+ * the use of this software.
  */
 
 #include <gtk/gtk.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <string.h>
 
-#include <audacious/gtk-compat.h>
 #include <audacious/i18n.h>
 #include <audacious/misc.h>
 #include <audacious/playlist.h>
 #include <libaudcore/audstrings.h>
 #include <libaudcore/hook.h>
 
-#include "config.h"
 #include "libaudgui.h"
 #include "libaudgui-gtk.h"
 
 #define AUDGUI_STATUS_TIMEOUT 3000
 #define IMAGE_SIZE 150
 
-static GtkWidget * infowin = NULL;
-
-static GtkWidget * location_text;
-static GtkWidget * entry_title;
-static GtkWidget * entry_artist;
-static GtkWidget * entry_album;
-static GtkWidget * entry_comment;
-static GtkWidget * entry_year;
-static GtkWidget * entry_track;
-static GtkWidget * entry_genre;
-
-static GtkWidget * image_artwork;
-
-static GtkWidget * label_format_name;
-static GtkWidget * label_quality;
-static GtkWidget * label_bitrate;
-static GtkWidget * btn_apply;
-static GtkWidget * label_mini_status;
-
 enum
 {
     RAWDATA_KEY,
@@ -63,9 +42,28 @@ enum
     RAWDATA_N_COLS
 };
 
+static GtkWidget * infowin = NULL;
+
+static struct {
+    GtkWidget * location;
+    GtkWidget * title;
+    GtkWidget * artist;
+    GtkWidget * album;
+    GtkWidget * comment;
+    GtkWidget * year;
+    GtkWidget * track;
+    GtkWidget * genre;
+    GtkWidget * image;
+    GtkWidget * format;
+    GtkWidget * quality;
+    GtkWidget * bitrate;
+    GtkWidget * apply;
+    GtkWidget * ministatus;
+} widgets;
+
 static char * current_file = NULL;
 static PluginHandle * current_decoder = NULL;
-static bool_t can_write = FALSE, something_changed = FALSE;
+static bool_t can_write = FALSE;
 
 /* This is by no means intended to be a complete list.  If it is not short, it
  * is useless: scrolling through ten pages of dropdown list is more work than
@@ -179,52 +177,10 @@ static void infowin_label_set_text (GtkWidget * widget, char * text)
     gtk_label_set_use_markup ((GtkLabel *) widget, TRUE);
 }
 
-static void infowin_entry_set_image (GtkWidget * widget, int list, int entry)
-{
-    GdkPixbuf * p = audgui_pixbuf_for_entry (list, entry);
-    g_return_if_fail (p);
-
-    audgui_pixbuf_scale_within (& p, IMAGE_SIZE);
-    gtk_image_set_from_pixbuf ((GtkImage *) widget, p);
-    g_object_unref ((GObject *) p);
-}
-
-static void clear_infowin (void)
-{
-    gtk_entry_set_text ((GtkEntry *) entry_title, "");
-    gtk_entry_set_text ((GtkEntry *) entry_artist, "");
-    gtk_entry_set_text ((GtkEntry *) entry_album, "");
-    gtk_entry_set_text ((GtkEntry *) entry_comment, "");
-    gtk_entry_set_text ((GtkEntry *) gtk_bin_get_child ((GtkBin *) entry_genre),
-     "");
-    gtk_entry_set_text ((GtkEntry *) entry_year, "");
-    gtk_entry_set_text ((GtkEntry *) entry_track, "");
-
-    infowin_label_set_text (label_format_name, NULL);
-    infowin_label_set_text (label_quality, NULL);
-    infowin_label_set_text (label_bitrate, NULL);
-
-    gtk_label_set_text ((GtkLabel *) label_mini_status,
-     "<span size=\"small\"></span>");
-    gtk_label_set_use_markup ((GtkLabel *) label_mini_status, TRUE);
-
-    g_free (current_file);
-    current_file = NULL;
-    current_decoder = NULL;
-
-    something_changed = FALSE;
-    can_write = FALSE;
-    gtk_widget_set_sensitive (btn_apply, FALSE);
-    gtk_image_clear ((GtkImage *) image_artwork);
-}
-
 static void entry_changed (GtkEditable * editable, void * unused)
 {
-    if (! something_changed && can_write)
-    {
-        something_changed = TRUE;
-        gtk_widget_set_sensitive (btn_apply, TRUE);
-    }
+    if (can_write)
+        gtk_widget_set_sensitive (widgets.apply, TRUE);
 }
 
 static bool_t ministatus_timeout_proc (void * data)
@@ -241,32 +197,31 @@ static void ministatus_display_message (const char * text)
 {
     char * tmp = g_strdup_printf ("<span size=\"small\">%s</span>", text);
 
-    gtk_label_set_text ((GtkLabel *) label_mini_status, tmp);
-    gtk_label_set_use_markup ((GtkLabel *) label_mini_status, TRUE);
+    gtk_label_set_text ((GtkLabel *) widgets.ministatus, tmp);
+    gtk_label_set_use_markup ((GtkLabel *) widgets.ministatus, TRUE);
     g_free (tmp);
 
     g_timeout_add (AUDGUI_STATUS_TIMEOUT, (GSourceFunc) ministatus_timeout_proc,
-     label_mini_status);
+     widgets.ministatus);
 }
 
 static void infowin_update_tuple (void * unused)
 {
     Tuple * tuple = tuple_new_from_filename (current_file);
 
-    set_field_str_from_entry (tuple, FIELD_TITLE, entry_title);
-    set_field_str_from_entry (tuple, FIELD_ARTIST, entry_artist);
-    set_field_str_from_entry (tuple, FIELD_ALBUM, entry_album);
-    set_field_str_from_entry (tuple, FIELD_COMMENT, entry_comment);
+    set_field_str_from_entry (tuple, FIELD_TITLE, widgets.title);
+    set_field_str_from_entry (tuple, FIELD_ARTIST, widgets.artist);
+    set_field_str_from_entry (tuple, FIELD_ALBUM, widgets.album);
+    set_field_str_from_entry (tuple, FIELD_COMMENT, widgets.comment);
     set_field_str_from_entry (tuple, FIELD_GENRE, gtk_bin_get_child ((GtkBin *)
-     entry_genre));
-    set_field_int_from_entry (tuple, FIELD_YEAR, entry_year);
-    set_field_int_from_entry (tuple, FIELD_TRACK_NUMBER, entry_track);
+     widgets.genre));
+    set_field_int_from_entry (tuple, FIELD_YEAR, widgets.year);
+    set_field_int_from_entry (tuple, FIELD_TRACK_NUMBER, widgets.track);
 
     if (aud_file_write_tuple (current_file, current_decoder, tuple))
     {
         ministatus_display_message (_("Metadata updated successfully"));
-        something_changed = FALSE;
-        gtk_widget_set_sensitive (btn_apply, FALSE);
+        gtk_widget_set_sensitive (widgets.apply, FALSE);
     }
     else
         ministatus_display_message (_("Metadata updating failed"));
@@ -292,6 +247,33 @@ static bool_t genre_fill (GtkWidget * combo)
     return FALSE;
 }
 
+static void infowin_display_image (const char * filename)
+{
+    if (! current_file || strcmp (filename, current_file))
+        return;
+
+    GdkPixbuf * pb = audgui_pixbuf_request (filename);
+    if (! pb)
+        pb = audgui_pixbuf_fallback ();
+
+    audgui_pixbuf_scale_within (& pb, IMAGE_SIZE);
+    gtk_image_set_from_pixbuf ((GtkImage *) widgets.image, pb);
+    g_object_unref (pb);
+}
+
+static void infowin_destroyed (void)
+{
+    hook_dissociate ("art ready", (HookFunction) infowin_display_image);
+
+    infowin = NULL;
+    memset (& widgets, 0, sizeof widgets);
+
+    str_unref (current_file);
+    current_file = NULL;
+    current_decoder = NULL;
+    can_write = FALSE;
+}
+
 static void create_infowin (void)
 {
     GtkWidget * hbox;
@@ -309,8 +291,8 @@ static void create_infowin (void)
     GtkWidget * label_quality_label;
     GtkWidget * label_bitrate_label;
     GtkWidget * codec_hbox;
-    GtkWidget * codec_table;
-    GtkWidget * table1;
+    GtkWidget * codec_grid;
+    GtkWidget * grid1;
     GtkWidget * bbox_close;
     GtkWidget * btn_close;
     GtkWidget * alignment;
@@ -321,33 +303,36 @@ static void create_infowin (void)
     gtk_window_set_type_hint ((GtkWindow *) infowin,
      GDK_WINDOW_TYPE_HINT_DIALOG);
 
-    vbox0 = gtk_vbox_new (FALSE, 0);
+    vbox0 = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
     gtk_container_add ((GtkContainer *) infowin, vbox0);
 
-    hbox = gtk_hbox_new (FALSE, 6);
+    hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL,  6);
     gtk_box_pack_start ((GtkBox *) vbox0, hbox, TRUE, TRUE, 0);
 
-    vbox2 = gtk_vbox_new (FALSE, 6);
+    vbox2 = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
     gtk_box_pack_start ((GtkBox *) hbox, vbox2, TRUE, TRUE, 0);
 
-    image_artwork = gtk_image_new ();
-    gtk_box_pack_start ((GtkBox *) vbox2, image_artwork, TRUE, TRUE, 0);
+    GtkWidget * vbox3 = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
+    gtk_box_pack_start ((GtkBox *) vbox2, vbox3, TRUE, FALSE, 0);
 
-    location_text = gtk_label_new ("");
-    gtk_widget_set_size_request (location_text, 200, -1);
-    gtk_label_set_line_wrap ((GtkLabel *) location_text, TRUE);
-    gtk_label_set_line_wrap_mode ((GtkLabel *) location_text,
-     PANGO_WRAP_WORD_CHAR);
-    gtk_label_set_selectable ((GtkLabel *) location_text, TRUE);
-    gtk_box_pack_start ((GtkBox *) vbox2, location_text, FALSE, FALSE, 0);
+    widgets.image = gtk_image_new ();
+    gtk_widget_set_size_request (widgets.image, IMAGE_SIZE, IMAGE_SIZE);
+    gtk_box_pack_start ((GtkBox *) vbox3, widgets.image, FALSE, FALSE, 0);
 
-    codec_hbox = gtk_hbox_new (FALSE, 6);
+    widgets.location = gtk_label_new ("");
+    gtk_widget_set_size_request (widgets.location, 200, -1);
+    gtk_label_set_line_wrap ((GtkLabel *) widgets.location, TRUE);
+    gtk_label_set_line_wrap_mode ((GtkLabel *) widgets.location, PANGO_WRAP_WORD_CHAR);
+    gtk_label_set_selectable ((GtkLabel *) widgets.location, TRUE);
+    gtk_box_pack_start ((GtkBox *) vbox3, widgets.location, FALSE, FALSE, 0);
+
+    codec_hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL,  6);
     gtk_box_pack_start ((GtkBox *) vbox2, codec_hbox, FALSE, FALSE, 0);
 
-    codec_table = gtk_table_new(3, 2, FALSE);
-    gtk_table_set_row_spacings ((GtkTable *) codec_table, 3);
-    gtk_table_set_col_spacings ((GtkTable *) codec_table, 12);
-    gtk_box_pack_start ((GtkBox *) codec_hbox, codec_table, FALSE, FALSE, 0);
+    codec_grid = gtk_grid_new ();
+    gtk_grid_set_row_spacing ((GtkGrid *) codec_grid, 3);
+    gtk_grid_set_column_spacing ((GtkGrid *) codec_grid, 12);
+    gtk_box_pack_start ((GtkBox *) codec_hbox, codec_grid, FALSE, FALSE, 0);
 
     label_format = gtk_label_new (_("<span size=\"small\">Format:</span>"));
     gtk_label_set_use_markup ((GtkLabel *) label_format, TRUE);
@@ -360,30 +345,24 @@ static void create_infowin (void)
     gtk_label_set_use_markup ((GtkLabel *) label_bitrate_label, TRUE);
     gtk_misc_set_alignment ((GtkMisc *) label_bitrate_label, 0, 0.5);
 
-    label_format_name = gtk_label_new (_("<span size=\"small\">n/a</span>"));
-    gtk_label_set_use_markup ((GtkLabel *) label_format_name, TRUE);
-    gtk_misc_set_alignment ((GtkMisc *) label_format_name, 0, 0.5);
-    label_quality = gtk_label_new (_("<span size=\"small\">n/a</span>"));
-    gtk_label_set_use_markup ((GtkLabel *) label_quality, TRUE);
-    gtk_misc_set_alignment ((GtkMisc *) label_quality, 0, 0.5);
-    label_bitrate = gtk_label_new (_("<span size=\"small\">n/a</span>"));
-    gtk_label_set_use_markup ((GtkLabel *) label_bitrate, TRUE);
-    gtk_misc_set_alignment ((GtkMisc *) label_bitrate, 0, 0.5);
+    widgets.format = gtk_label_new (_("<span size=\"small\">n/a</span>"));
+    gtk_label_set_use_markup ((GtkLabel *) widgets.format, TRUE);
+    gtk_misc_set_alignment ((GtkMisc *) widgets.format, 0, 0.5);
+    widgets.quality = gtk_label_new (_("<span size=\"small\">n/a</span>"));
+    gtk_label_set_use_markup ((GtkLabel *) widgets.quality, TRUE);
+    gtk_misc_set_alignment ((GtkMisc *) widgets.quality, 0, 0.5);
+    widgets.bitrate = gtk_label_new (_("<span size=\"small\">n/a</span>"));
+    gtk_label_set_use_markup ((GtkLabel *) widgets.bitrate, TRUE);
+    gtk_misc_set_alignment ((GtkMisc *) widgets.bitrate, 0, 0.5);
 
-    gtk_table_attach ((GtkTable *) codec_table, label_format, 0, 1, 0, 1,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
-    gtk_table_attach ((GtkTable *) codec_table, label_format_name, 1, 2, 0, 1,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
-    gtk_table_attach ((GtkTable *) codec_table, label_quality_label, 0, 1, 1, 2,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
-    gtk_table_attach ((GtkTable *) codec_table, label_quality, 1, 2, 1, 2,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
-    gtk_table_attach ((GtkTable *) codec_table, label_bitrate_label, 0, 1, 2, 3,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
-    gtk_table_attach ((GtkTable *) codec_table, label_bitrate, 1, 2, 2, 3,
-     GTK_EXPAND | GTK_FILL, 0, 0, 0);
+    gtk_grid_attach ((GtkGrid *) codec_grid, label_format, 0, 0, 1, 1);
+    gtk_grid_attach ((GtkGrid *) codec_grid, widgets.format, 1, 0, 1, 1);
+    gtk_grid_attach ((GtkGrid *) codec_grid, label_quality_label, 0, 1, 1, 1);
+    gtk_grid_attach ((GtkGrid *) codec_grid, widgets.quality, 1, 1, 1, 1);
+    gtk_grid_attach ((GtkGrid *) codec_grid, label_bitrate_label, 0, 2, 1, 1);
+    gtk_grid_attach ((GtkGrid *) codec_grid, widgets.bitrate, 1, 2, 1, 1);
 
-    vbox2 = gtk_vbox_new (FALSE, 0);
+    vbox2 = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
     gtk_box_pack_start ((GtkBox *) hbox, vbox2, TRUE, TRUE, 0);
 
     label_title = gtk_label_new (_("<span size=\"small\">Title</span>"));
@@ -394,9 +373,9 @@ static void create_infowin (void)
     alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
-    entry_title = gtk_entry_new ();
-    gtk_container_add ((GtkContainer *) alignment, entry_title);
-    g_signal_connect (entry_title, "changed", (GCallback) entry_changed, NULL);
+    widgets.title = gtk_entry_new ();
+    gtk_container_add ((GtkContainer *) alignment, widgets.title);
+    g_signal_connect (widgets.title, "changed", (GCallback) entry_changed, NULL);
 
     label_artist = gtk_label_new (_("<span size=\"small\">Artist</span>"));
     gtk_box_pack_start ((GtkBox *) vbox2, label_artist, FALSE, FALSE, 0);
@@ -406,9 +385,9 @@ static void create_infowin (void)
     alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
-    entry_artist = gtk_entry_new ();
-    gtk_container_add ((GtkContainer *) alignment, entry_artist);
-    g_signal_connect (entry_artist, "changed", (GCallback) entry_changed, NULL);
+    widgets.artist = gtk_entry_new ();
+    gtk_container_add ((GtkContainer *) alignment, widgets.artist);
+    g_signal_connect (widgets.artist, "changed", (GCallback) entry_changed, NULL);
 
     label_album = gtk_label_new (_("<span size=\"small\">Album</span>"));
     gtk_box_pack_start ((GtkBox *) vbox2, label_album, FALSE, FALSE, 0);
@@ -418,9 +397,9 @@ static void create_infowin (void)
     alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
-    entry_album = gtk_entry_new ();
-    gtk_container_add ((GtkContainer *) alignment, entry_album);
-    g_signal_connect (entry_album, "changed", (GCallback) entry_changed, NULL);
+    widgets.album = gtk_entry_new ();
+    gtk_container_add ((GtkContainer *) alignment, widgets.album);
+    g_signal_connect (widgets.album, "changed", (GCallback) entry_changed, NULL);
 
     label_comment = gtk_label_new (_("<span size=\"small\">Comment</span>"));
     gtk_box_pack_start ((GtkBox *) vbox2, label_comment, FALSE, FALSE, 0);
@@ -430,9 +409,9 @@ static void create_infowin (void)
     alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
-    entry_comment = gtk_entry_new ();
-    gtk_container_add ((GtkContainer *) alignment, entry_comment);
-    g_signal_connect (entry_comment, "changed", (GCallback) entry_changed, NULL);
+    widgets.comment = gtk_entry_new ();
+    gtk_container_add ((GtkContainer *) alignment, widgets.comment);
+    g_signal_connect (widgets.comment, "changed", (GCallback) entry_changed, NULL);
 
     label_genre = gtk_label_new (_("<span size=\"small\">Genre</span>"));
     gtk_box_pack_start ((GtkBox *) vbox2, label_genre, FALSE, FALSE, 0);
@@ -443,71 +422,69 @@ static void create_infowin (void)
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
 
-    entry_genre = gtk_combo_box_text_new_with_entry ();
-    gtk_container_add ((GtkContainer *) alignment, entry_genre);
-    g_signal_connect (entry_genre, "changed", (GCallback) entry_changed, NULL);
-    g_idle_add ((GSourceFunc) genre_fill, entry_genre);
+    widgets.genre = gtk_combo_box_text_new_with_entry ();
+    gtk_container_add ((GtkContainer *) alignment, widgets.genre);
+    g_signal_connect (widgets.genre, "changed", (GCallback) entry_changed, NULL);
+    g_idle_add ((GSourceFunc) genre_fill, widgets.genre);
 
     alignment = gtk_alignment_new (0.5, 0.5, 1, 1);
     gtk_box_pack_start ((GtkBox *) vbox2, alignment, FALSE, FALSE, 0);
     gtk_alignment_set_padding ((GtkAlignment *) alignment, 0, 6, 0, 0);
-    table1 = gtk_table_new (2, 2, FALSE);
-    gtk_container_add ((GtkContainer *) alignment, table1);
-    gtk_table_set_col_spacings ((GtkTable *) table1, 6);
+    grid1 = gtk_grid_new ();
+    gtk_grid_set_column_homogeneous ((GtkGrid *) grid1, TRUE);
+    gtk_container_add ((GtkContainer *) alignment, grid1);
+    gtk_grid_set_column_spacing ((GtkGrid *) grid1, 6);
 
     label_year = gtk_label_new (_("<span size=\"small\">Year</span>"));
-    gtk_table_attach ((GtkTable *) table1, label_year, 0, 1, 0, 1, GTK_FILL, 0,
-     0, 0);
+    gtk_grid_attach ((GtkGrid *) grid1, label_year, 0, 0, 1, 1);
     gtk_label_set_use_markup ((GtkLabel *) label_year, TRUE);
     gtk_misc_set_alignment ((GtkMisc *) label_year, 0, 0.5);
 
-    entry_year = gtk_entry_new ();
-    gtk_table_attach ((GtkTable *) table1, entry_year, 0, 1, 1, 2, GTK_EXPAND |
-     GTK_FILL, 0, 0, 0);
-    g_signal_connect (entry_year, "changed", (GCallback) entry_changed, NULL);
+    widgets.year = gtk_entry_new ();
+    gtk_grid_attach ((GtkGrid *) grid1, widgets.year, 0, 1, 1, 1);
+    g_signal_connect (widgets.year, "changed", (GCallback) entry_changed, NULL);
 
     label_track = gtk_label_new (_("<span size=\"small\">Track Number</span>"));
-    gtk_table_attach ((GtkTable *) table1, label_track, 1, 2, 0, 1, GTK_FILL, 0,
-     0, 0);
+    gtk_grid_attach ((GtkGrid *) grid1, label_track, 1, 0, 1, 1);
     gtk_label_set_use_markup ((GtkLabel *) label_track, TRUE);
     gtk_misc_set_alignment ((GtkMisc *) label_track, 0, 0.5);
 
-    entry_track = gtk_entry_new ();
-    gtk_table_attach ((GtkTable *) table1, entry_track, 1, 2, 1, 2, GTK_EXPAND |
-     GTK_FILL, 0, 0, 0);
-    g_signal_connect (entry_track, "changed", (GCallback) entry_changed, NULL);
+    widgets.track = gtk_entry_new ();
+    gtk_grid_attach ((GtkGrid *) grid1, widgets.track, 1, 1, 1, 1);
+    g_signal_connect (widgets.track, "changed", (GCallback) entry_changed, NULL);
 
-    hbox_status_and_bbox = gtk_hbox_new (FALSE, 0);
+    hbox_status_and_bbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL,  0);
     gtk_box_pack_start ((GtkBox *) vbox0, hbox_status_and_bbox, FALSE, FALSE, 0);
 
-    label_mini_status = gtk_label_new ("<span size=\"small\"></span>");
-    gtk_label_set_use_markup ((GtkLabel *) label_mini_status, TRUE);
-    gtk_misc_set_alignment ((GtkMisc *) label_mini_status, 0, 0.5);
-    gtk_box_pack_start ((GtkBox *) hbox_status_and_bbox, label_mini_status,
+    widgets.ministatus = gtk_label_new ("<span size=\"small\"></span>");
+    gtk_label_set_use_markup ((GtkLabel *) widgets.ministatus, TRUE);
+    gtk_misc_set_alignment ((GtkMisc *) widgets.ministatus, 0, 0.5);
+    gtk_box_pack_start ((GtkBox *) hbox_status_and_bbox, widgets.ministatus,
      TRUE, TRUE, 0);
 
-    bbox_close = gtk_hbutton_box_new ();
+    bbox_close = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
     gtk_box_set_spacing ((GtkBox *) bbox_close, 6);
     gtk_box_pack_start ((GtkBox *) hbox_status_and_bbox, bbox_close, FALSE,
      FALSE, 0);
     gtk_button_box_set_layout ((GtkButtonBox *) bbox_close, GTK_BUTTONBOX_END);
 
-    btn_apply = gtk_button_new_from_stock (GTK_STOCK_SAVE);
-    gtk_container_add ((GtkContainer *) bbox_close, btn_apply);
-    g_signal_connect (btn_apply, "clicked", (GCallback) infowin_update_tuple,
-     NULL);
-    gtk_widget_set_sensitive (btn_apply, FALSE);
+    widgets.apply = gtk_button_new_from_stock (GTK_STOCK_SAVE);
+    gtk_container_add ((GtkContainer *) bbox_close, widgets.apply);
+    g_signal_connect (widgets.apply, "clicked", (GCallback) infowin_update_tuple, NULL);
+    gtk_widget_set_sensitive (widgets.apply, FALSE);
 
     btn_close = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
     gtk_container_add ((GtkContainer *) bbox_close, btn_close);
     g_signal_connect_swapped (btn_close, "clicked", (GCallback) gtk_widget_hide,
      infowin);
 
-    audgui_hide_on_delete (infowin);
-    audgui_hide_on_escape (infowin);
-
     gtk_widget_show_all (vbox0);
-    gtk_widget_grab_focus (entry_title);
+    gtk_widget_grab_focus (widgets.title);
+
+    audgui_destroy_on_escape (infowin);
+    g_signal_connect (infowin, "destroy", (GCallback) infowin_destroyed, NULL);
+
+    hook_associate ("art ready", (HookFunction) infowin_display_image, NULL);
 }
 
 static void infowin_show (int list, int entry, const char * filename,
@@ -515,42 +492,38 @@ static void infowin_show (int list, int entry, const char * filename,
 {
     char * tmp;
 
-    if (infowin == NULL)
+    if (! infowin)
         create_infowin ();
-    else
-        clear_infowin ();
 
-    current_file = g_strdup (filename);
+    str_unref (current_file);
+    current_file = str_get (filename);
     current_decoder = decoder;
     can_write = updating_enabled;
 
-    set_entry_str_from_field (entry_title, tuple, FIELD_TITLE, updating_enabled);
-    set_entry_str_from_field (entry_artist, tuple, FIELD_ARTIST,
-     updating_enabled);
-    set_entry_str_from_field (entry_album, tuple, FIELD_ALBUM, updating_enabled);
-    set_entry_str_from_field (entry_comment, tuple, FIELD_COMMENT,
-     updating_enabled);
-    set_entry_str_from_field (gtk_bin_get_child ((GtkBin *) entry_genre), tuple,
-     FIELD_GENRE, updating_enabled);
+    set_entry_str_from_field (widgets.title, tuple, FIELD_TITLE, updating_enabled);
+    set_entry_str_from_field (widgets.artist, tuple, FIELD_ARTIST, updating_enabled);
+    set_entry_str_from_field (widgets.album, tuple, FIELD_ALBUM, updating_enabled);
+    set_entry_str_from_field (widgets.comment, tuple, FIELD_COMMENT, updating_enabled);
+    set_entry_str_from_field (gtk_bin_get_child ((GtkBin *) widgets.genre),
+     tuple, FIELD_GENRE, updating_enabled);
 
     tmp = uri_to_display (filename);
-    gtk_label_set_text ((GtkLabel *) location_text, tmp);
+    gtk_label_set_text ((GtkLabel *) widgets.location, tmp);
     g_free (tmp);
 
-    set_entry_int_from_field (entry_year, tuple, FIELD_YEAR, updating_enabled);
-    set_entry_int_from_field (entry_track, tuple, FIELD_TRACK_NUMBER,
-     updating_enabled);
+    set_entry_int_from_field (widgets.year, tuple, FIELD_YEAR, updating_enabled);
+    set_entry_int_from_field (widgets.track, tuple, FIELD_TRACK_NUMBER, updating_enabled);
 
-    infowin_label_set_text (label_format_name, tuple_get_str (tuple, FIELD_CODEC, NULL));
-    infowin_label_set_text (label_quality, tuple_get_str (tuple, FIELD_QUALITY, NULL));
+    infowin_label_set_text (widgets.format, tuple_get_str (tuple, FIELD_CODEC, NULL));
+    infowin_label_set_text (widgets.quality, tuple_get_str (tuple, FIELD_QUALITY, NULL));
 
     if (tuple_get_value_type (tuple, FIELD_BITRATE, NULL) == TUPLE_INT)
-        infowin_label_set_text (label_bitrate, str_printf (_("%d kb/s"),
+        infowin_label_set_text (widgets.bitrate, str_printf (_("%d kb/s"),
          tuple_get_int (tuple, FIELD_BITRATE, NULL)));
     else
-        infowin_label_set_text (label_bitrate, NULL);
+        infowin_label_set_text (widgets.bitrate, NULL);
 
-    infowin_entry_set_image (image_artwork, list, entry);
+    infowin_display_image (filename);
 
     gtk_window_present ((GtkWindow *) infowin);
 }
@@ -602,3 +575,9 @@ EXPORT void audgui_infowin_show_current (void)
 
     audgui_infowin_show (playlist, position);
 }
+
+EXPORT void audgui_infowin_hide (void)
+{
+    if (infowin)
+        gtk_widget_destroy (infowin);
+}
