@@ -1,22 +1,21 @@
 /*
- * Copyright 2009 Paula Stanciu
- * Copyright 2010 John Lindgren
+ * id3v24.c
+ * Copyright 2009-2011 Paula Stanciu, Tony Vroon, John Lindgren,
+ *                     Mikael Magnusson, and Michał Lipski
  *
- * This file is part of Audacious.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * Audacious is free software: you can redistribute it and/or modify it under
- * the terms of the GNU General Public License as published by the Free Software
- * Foundation, version 3 of the License.
+ * 1. Redistributions of source code must retain the above copyright notice,
+ *    this list of conditions, and the following disclaimer.
  *
- * Audacious is distributed in the hope that it will be useful, but WITHOUT ANY
- * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
- * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
+ * 2. Redistributions in binary form must reproduce the above copyright notice,
+ *    this list of conditions, and the following disclaimer in the documentation
+ *    provided with the distribution.
  *
- * You should have received a copy of the GNU General Public License along with
- * Audacious. If not, see <http://www.gnu.org/licenses/>.
- *
- * The Audacious team does not consider modular code linking to Audacious or
- * using our public API to be a derived work.
+ * This software is provided "as is" and without any warranty, express or
+ * implied. In no event shall the authors be liable for any damages arising from
+ * the use of this software.
  */
 
 #include <glib.h>
@@ -37,7 +36,6 @@ enum
     ID3_COMPOSER,
     ID3_COPYRIGHT,
     ID3_DATE,
-    ID3_TIME,
     ID3_LENGTH,
     ID3_ARTIST,
     ID3_TRACKNR,
@@ -52,9 +50,9 @@ enum
     ID3_TAGS_NO
 };
 
-static const char * id3_frames[ID3_TAGS_NO] = {"TALB","TIT2","TCOM", "TCOP",
- "TDAT", "TIME", "TLEN", "TPE1", "TRCK", "TYER","TCON", "COMM", "PRIV", "TSSE",
- "TDRC", "TXXX", "RVA2"};
+static const char * id3_frames[ID3_TAGS_NO] = {"TALB", "TIT2", "TCOM", "TCOP",
+ "TDAT", "TLEN", "TPE1", "TRCK", "TYER", "TCON", "COMM", "PRIV", "TSSE", "TDRC",
+ "TXXX", "RVA2"};
 
 static const unsigned char PRIMARY_CLASS_MUSIC[16] = {0xBC, 0x7D, 0x60, 0xD1, 0x23,
  0xE3, 0xE2, 0x4B, 0x86, 0xA1, 0x48, 0xA4, 0x2A, 0x28, 0x44, 0x1E};
@@ -318,7 +316,7 @@ static bool_t read_frame (VFSFile * handle, int max_size, int version,
     TAGDBG (" flags = %x\n", (int) header.flags);
 
     * frame_size = sizeof (ID3v2FrameHeader) + header.size;
-    sprintf (key, "%.4s", header.key);
+    g_strlcpy (key, header.key, 5);
 
     if (header.flags & (ID3_FRAME_COMPRESSED | ID3_FRAME_ENCRYPTED))
     {
@@ -860,9 +858,6 @@ static bool_t id3v24_read_tag (Tuple * tuple, VFSFile * handle)
           case ID3_DATE:
             associate_string (tuple, FIELD_DATE, NULL, data, size);
             break;
-          case ID3_TIME:
-            associate_int (tuple, FIELD_LENGTH, NULL, data, size);
-            break;
           case ID3_LENGTH:
             associate_int (tuple, FIELD_LENGTH, NULL, data, size);
             break;
